! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: edetaux_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some utilities for patching, converting and generally manipulating source lists created by the standard pcms-type source detection tasks into formats acceptable to the eimsim script.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module edetaux_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.

  use dal
  use param  
  use errorHandling

!  use temp_ssclib

  use array_utils	! smallestTrueIndex().
  use constants
  use coordinate	! for raDecToTan().
  use dal_aux		! addOrOpenColumn().
!use l_dal_aux_2		! addOrOpenColumn().
  use wcs_aux

  implicit none 

  private 

  ! Command-line parameters:
  character(20) ::&
    functionType=''
  character(256) ::&
    inSetName='',&
    outSetName=''

  public edetaux

contains 

  !---------------------------------------------------------------------
  subroutine edetaux
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    select case(functionType)
    case('convert')
      call convertFromSrcmatchToEmldetect
    case('addboxid')
      call addBoxId
    case('sky2pix')
      call sky2Pix
    case('addboxcols')
      call srcListPatch
    case default
call error('',"")
    end select

  end subroutine edetaux

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

    functionType = stringParameter('function')
    inSetName = stringParameter('inset')
    outSetName = stringParameter('outset')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine convertFromSrcmatchToEmldetect
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) :: srcmatchSet, emldetectSet
    type(TableT) :: srcmatchTab, emldetectTab
    integer ::&
      numSources, idInst, idBand, numInstsBandsPerSource,&
      numRowsEmldetectList, srcNum, startRow, finisRow, firstNonInst0Row,&
      row, instFinisOffset, instStartOffset, rowOffset
    character(15) :: oldColName=''
    character(2) :: instrStr(3) = (/'PN','M1','M2'/)
    logical :: instrBandPresent(0:3, 0:9) = .false.
    type(ColumnT) ::&
      srcIdCol, mlIdCol, boxIdCol, idInstCol, idBandCol, idClusterCol,&
      sctsCol, sctsErrCol, ximaCol, ximaErrCol, yimaCol, yimaErrCol,&
      extCol, extErrCol, detmlCol, extmlCol, bgMapCol, expMapCol, fluxCol,&
      fluxErrCol, rateCol, rateErrCol, raCol, decCol, raDecErrCol, liiCol,&
      biiCol, rawXCol, rawYCol, ccdnrCol, hr1Col, hr1ErrCol, hr2Col,&
      hr2ErrCol, hr3Col, hr3ErrCol, cutRadCol, maskFracCol, eefCol,&
      vignetCol, distNNCol, valFlagCol, verFlagCol, verCommCol,&
      oldValFlagCol, oldVerFlagCol, oldVerCommCol
    integer(int32), pointer, dimension(:) ::&
      srcId, oldSrcId, mlId, oldMlId, boxId, oldBoxId, idInstData, idBandData
    real(single), pointer, dimension(:) ::&
      ext, oldExt, extErr, oldExtErr, detml, oldDetml, extml, oldExtml,&
      expMap, oldExpMap, flux, oldFlux, fluxErr, oldFluxErr, rate, oldRate,&
      rateErr, oldRateErr, raDecErr, oldRaDecErr, hr1, oldHr1, hr1Err,&
      oldHr1Err, hr2, oldHr2, hr2Err, oldHr2Err, hr3, oldHr3, hr3Err,&
      oldHr3Err
    real(double), pointer, dimension(:) ::&
      ra, oldRa, dec, oldDec, lii, oldLii, bii, oldBii
    character(40) :: emptyStr=''
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'aaa'
    srcmatchSet = dataSet(inSetName, READ)
    srcmatchTab = table(srcmatchSet, 'SRCLIST')
    emldetectSet = dataSet(outSetName, CREATE)
    call setAttributes(attributable(emldetectSet), attributable(srcmatchSet))

    !...................................................................
    ! First find out how many sources, instruments and bands there are:

    numSources = numberOfRows(srcmatchTab)
    do idInst = 1, 3
      do idBand = 1, 9
        write(oldColName, '(a3, i1, a5)')&
          instrStr(idInst)//'_', idBand, '_FLUX'
        if (hasColumn(srcmatchTab, oldColName))&
          instrBandPresent(idInst, idBand) = .true.
      end do
      if (count(instrBandPresent(idInst, 1:9)) > 1)&
      ! then should also have a summary band row:
        instrBandPresent(idInst, 0) = .true.
    end do
    if (count(instrBandPresent(1:3, 0)) > 1)&
    ! then should also have a summary inst row:
      instrBandPresent(0, 0) = .true.

    numInstsBandsPerSource = count(instrBandPresent)
    numRowsEmldetectList = numSources * numInstsBandsPerSource
    emldetectTab = addTable(emldetectSet, 'SRCLIST', numRowsEmldetectList)
    call setAttributes(attributable(emldetectTab), attributable(srcmatchTab))

    !...................................................................
    ! Now add all the necessary columns for the output list:

    srcIdCol     = addColumn(emldetectTab, 'SRC_NUM',    INTEGER32)
    mlIdCol      = addColumn(emldetectTab, 'ML_ID_SRC',  INTEGER32)
    boxIdCol     = addColumn(emldetectTab, 'BOX_ID_SRC', INTEGER32)
    idInstCol    = addColumn(emldetectTab, 'ID_INST',    INTEGER32)
    idBandCol    = addColumn(emldetectTab, 'ID_BAND',    INTEGER32)
    idClusterCol = addColumn(emldetectTab, 'ID_CLUSTER', INTEGER32)
    sctsCol      = addColumn(emldetectTab, 'SCTS',       REAL32&
      , 'counts')
    sctsErrCol   = addColumn(emldetectTab, 'SCTS_ERR',   REAL32&
      , 'counts')
    ximaCol      = addColumn(emldetectTab, 'X_IMA',      REAL32&
      , 'image pixels')
    ximaErrCol   = addColumn(emldetectTab, 'X_IMA_ERR',  REAL32&
      , 'image pixels')
    yimaCol      = addColumn(emldetectTab, 'Y_IMA',      REAL32&
      , 'image pixels')
    yimaErrCol   = addColumn(emldetectTab, 'Y_IMA_ERR',  REAL32&
      , 'image pixels')
    extCol       = addColumn(emldetectTab, 'EXT',        REAL32&
      , 'image pixels')
    extErrCol    = addColumn(emldetectTab, 'EXT_ERR',    REAL32&
      , 'image pixels')
    detmlCol     = addColumn(emldetectTab, 'DET_ML',     REAL32)
    extmlCol     = addColumn(emldetectTab, 'EXT_ML',     REAL32)
    bgMapCol     = addColumn(emldetectTab, 'BG_MAP',     REAL32&
      , 'counts/pixel')
    expMapCol    = addColumn(emldetectTab, 'EXP_MAP',    REAL32&
      , 's')
    fluxCol      = addColumn(emldetectTab, 'FLUX',       REAL32&
      , 'cgs')
    fluxErrCol   = addColumn(emldetectTab, 'FLUX_ERR',   REAL32&
      , 'cgs')
    rateCol      = addColumn(emldetectTab, 'RATE',       REAL32&
      , 'counts/s')
    rateErrCol   = addColumn(emldetectTab, 'RATE_ERR',   REAL32&
      , 'counts/s')
    raCol        = addColumn(emldetectTab, 'RA',         REAL64&
      , 'deg')
    decCol       = addColumn(emldetectTab, 'DEC',        REAL64&
      , 'deg')
    raDecErrCol  = addColumn(emldetectTab, 'RADEC_ERR',  REAL32&
      , 'arcsec')
    liiCol       = addColumn(emldetectTab, 'LII',        REAL64&
      , 'deg')
    biiCol       = addColumn(emldetectTab, 'BII',        REAL64&
      , 'deg')
    rawXCol      = addColumn(emldetectTab, 'RAWX',       INTEGER32&
      , 'ccd pixels')
    rawYCol      = addColumn(emldetectTab, 'RAWY',       INTEGER32&
      , 'ccd pixels')
    ccdnrCol     = addColumn(emldetectTab, 'CCDNR',      INTEGER32)
    hr1Col       = addColumn(emldetectTab, 'HR1',        REAL32)
    hr1ErrCol    = addColumn(emldetectTab, 'HR1_ERR',    REAL32)
    hr2Col       = addColumn(emldetectTab, 'HR2',        REAL32)
    hr2ErrCol    = addColumn(emldetectTab, 'HR2_ERR',    REAL32)
    hr3Col       = addColumn(emldetectTab, 'HR3',        REAL32)
    hr3ErrCol    = addColumn(emldetectTab, 'HR3_ERR',    REAL32)
    cutRadCol    = addColumn(emldetectTab, 'CUTRAD',     REAL32&
      , 'image pixels')
    maskFracCol  = addColumn(emldetectTab, 'MASKFRAC',   REAL32)
    eefCol       = addColumn(emldetectTab, 'EEF',        REAL32)
    vignetCol    = addColumn(emldetectTab, 'VIGNETTING', REAL32)
    distNNCol    = addColumn(emldetectTab, 'DIST_NN',    REAL32&
      , 'arcsec')
    valFlagCol   = addColumn(emldetectTab, 'VAL_FLAG',   STRING&
      , dimensions=(/12/))
    verFlagCol   = addColumn(emldetectTab, 'VER_FLAG',   STRING&
      , dimensions=(/12/))
    verCommCol   = addColumn(emldetectTab, 'VER_COMM',   STRING&
      , dimensions=(/40/))

    !...................................................................
    ! Now work through the output columns and insert the correct data into them:
if (testFlag) print *, 'bbb'

    ! SRC_NUM
if (testFlag) print *, 'SRC_NUM'

    srcId => int32Data(srcIdCol)
    oldSrcId => int32Data(column(srcmatchTab, 'SRC_NUM'))
    do srcNum = 1, numSources
      startRow = numInstsBandsPerSource * (srcNum-1)
      finisRow = numInstsBandsPerSource * srcNum - 1
      srcId(startRow:finisRow) = oldSrcId(srcNum-1)
    end do
    call release(srcIdCol)

    ! ML_ID_SRC
if (testFlag) print *, 'ML_ID_SRC'

    ! First deal with rows for which ID_INST==0:
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      firstNonInst0Row = 1
      call setNullValue(mlIdCol, -1)
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        call setToNull(mlIdCol, row)
      end do
    end if

    mlId => int32Data(mlIdCol)
    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
!print *, ''
!print *, 'inst=', idInst
      oldColName = instrStr(idInst)//'_ML_ID_SRC'
      oldMlId => int32Data(column(srcmatchTab, oldColName))
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
!print *, '  ', srcNum, startRow, finisRow, oldMlId(srcNum-1)
        mlId(startRow:finisRow) = oldMlId(srcNum-1)
      end do
    end do
    call release(mlIdCol)

    ! BOX_ID_SRC
if (testFlag) print *, 'BOX_ID_SRC'

    ! First deal with rows for which ID_INST==0:
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      firstNonInst0Row = 1
      call setNullValue(boxIdCol, -1)
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        call setToNull(boxIdCol, row)
      end do
    end if

    boxId => int32Data(boxIdCol)
    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      oldColName = instrStr(idInst)//'_BOX_ID_SRC'
      oldBoxId => int32Data(column(srcmatchTab, oldColName))
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        boxId(startRow:finisRow) = oldBoxId(srcNum-1)
      end do
    end do
    call release(boxIdCol)

    ! ID_INST
if (testFlag) print *, 'ID_INST'

    idInstData => int32Data(idInstCol)
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      firstNonInst0Row = 1
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        idInstData(row) = 0
      end do
    end if

    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        idInstData(startRow:finisRow) = idInst
      end do
    end do
    call release(idInstCol)

    ! ID_BAND
if (testFlag) print *, 'ID_BAND'

    idBandData => int32Data(idBandCol)
    rowOffset = 0
    if (instrBandPresent(0, 0)) then
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1) + rowOffset
        idBandData(row) = 0
      end do
      rowOffset = rowOffset + 1
    end if

    do idInst = 1, 3
      if (instrBandPresent(idInst, 0)) then
        do srcNum = 1, numSources
          row = numInstsBandsPerSource * (srcNum-1) + rowOffset
          idBandData(row) = 0
        end do
        rowOffset = rowOffset + 1
      end if

      do idBand = 1, 9
      if (.not. instrBandPresent(idInst, idBand)) cycle
        do srcNum = 1, numSources
          row = numInstsBandsPerSource * (srcNum-1) + rowOffset
          idBandData(row) = idBand
        end do
        rowOffset = rowOffset + 1
      end do ! idband
    end do ! idInst
    call release(idBandCol)

    ! ID_CLUSTER
if (testFlag) print *, 'ID_CLUSTER'

    call setNullValue(idClusterCol, -1)
    do row = 0, numRowsEmldetectList-1
      call setToNull(idClusterCol, row)
    end do
    call release(idClusterCol)

    ! SCTS
if (testFlag) print *, 'SCTS'

!    call setNullValue(sctsCol, -1)
    do row = 0, numRowsEmldetectList-1
      call setToNull(sctsCol, row)
    end do
    call release(sctsCol)

    ! SCTS_ERR
if (testFlag) print *, 'SCTS_ERR'

!    call setNullValue(sctsErrCol, -1)
    do row = 0, numRowsEmldetectList-1
      call setToNull(sctsErrCol, row)
    end do
    call release(sctsErrCol)

    ! X_IMA
if (testFlag) print *, 'X_IMA'

!    call setNullValue(ximaCol, -1)
    do row = 0, numRowsEmldetectList-1
      call setToNull(ximaCol, row)
    end do
    call release(ximaCol)

    ! X_IMA_ERR
if (testFlag) print *, 'X_IMA_ERR'

!    call setNullValue(ximaErrCol, -1)
    do row = 0, numRowsEmldetectList-1
      call setToNull(ximaErrCol, row)
    end do
    call release(ximaErrCol)

    ! Y_IMA
if (testFlag) print *, 'Y_IMA'

!    call setNullValue(yimaCol, -1)
    do row = 0, numRowsEmldetectList-1
      call setToNull(yimaCol, row)
    end do
    call release(yimaCol)

    ! Y_IMA_ERR
if (testFlag) print *, 'Y_IMA_ERR'

!    call setNullValue(yimaErrCol, -1)
    do row = 0, numRowsEmldetectList-1
      call setToNull(yimaErrCol, row)
    end do
    call release(yimaErrCol)

    ! EXT
if (testFlag) print *, 'EXT'

    ! First deal with rows for which ID_INST==0:
    ext => real32Data(extCol)
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      oldColName = 'EP_EXTENT'
      oldExt => real32Data(column(srcmatchTab, oldColName))
      firstNonInst0Row = 1
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        ext(row) = oldExt(srcNum-1)
      end do
    end if

    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      oldColName = instrStr(idInst)//'_EXTENT'
      oldExt => real32Data(column(srcmatchTab, oldColName))
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        ext(startRow:finisRow) = oldExt(srcNum-1)
      end do
    end do
    call release(extCol)

    ! EXT_ERR
if (testFlag) print *, 'EXT_ERR'

    ! First deal with rows for which ID_INST==0:
    extErr => real32Data(extErrCol)
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      oldColName = 'EP_EXT_ERR'
      oldExtErr => real32Data(column(srcmatchTab, oldColName))
      firstNonInst0Row = 1
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        extErr(row) = oldExtErr(srcNum-1)
      end do
    end if

    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      oldColName = instrStr(idInst)//'_EXT_ERR'
      oldExt => real32Data(column(srcmatchTab, oldColName))
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        extErr(startRow:finisRow) = oldExtErr(srcNum-1)
      end do
    end do
    call release(extErrCol)

    ! DET_ML
if (testFlag) print *, 'DET_ML'

    ! First deal with rows for which ID_INST==0:
    detml => real32Data(detmlCol)
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      oldColName = 'EP_DET_ML'
      oldDetml => real32Data(column(srcmatchTab, oldColName))
      firstNonInst0Row = 1
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        detml(row) = oldDetml(srcNum-1)
      end do
    end if

    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      oldColName = instrStr(idInst)//'_DET_ML'
      oldDetml => real32Data(column(srcmatchTab, oldColName))
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        detml(startRow:finisRow) = oldDetml(srcNum-1)
      end do
    end do
    call release(detmlCol)

    ! EXT_ML
if (testFlag) print *, 'EXT_ML'

    ! First deal with rows for which ID_INST==0:
    extml => real32Data(extmlCol)
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      oldColName = 'EP_EXT_ML'
      oldExtml => real32Data(column(srcmatchTab, oldColName))
      firstNonInst0Row = 1
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        extml(row) = oldExtml(srcNum-1)
      end do
    end if

    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      oldColName = instrStr(idInst)//'_EXT_ML'
      oldExtml => real32Data(column(srcmatchTab, oldColName))
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        extml(startRow:finisRow) = oldExtml(srcNum-1)
      end do
    end do
    call release(extmlCol)

    ! BG_MAP
if (testFlag) print *, 'BG_MAP'

    do row = 0, numRowsEmldetectList-1
      call setToNull(bgMapCol, row)
    end do
    call release(bgMapCol)

    ! EXP_MAP
if (testFlag) print *, 'EXP_MAP'

    ! First deal with rows for which ID_INST==0:
    rowOffset = 0
    if (instrBandPresent(0, 0)) then
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1) + rowOffset
        call setToNull(expMapCol, row)
      end do
      rowOffset = rowOffset + 1
    end if

    do idInst = 1, 3
      if (instrBandPresent(idInst, 0)) then
        do srcNum = 1, numSources
          row = numInstsBandsPerSource * (srcNum-1) + rowOffset
          call setToNull(expMapCol, row)
        end do
        rowOffset = rowOffset + 1
      end if
      rowOffset = rowOffset + count(instrBandPresent(idInst, 1:9))
    end do

    rowOffset = 0
    if (instrBandPresent(0, 0)) then
      rowOffset = rowOffset + 1
    end if

    expMap => real32Data(expMapCol)
    do idInst = 1, 3
      if (instrBandPresent(idInst, 0)) then
        rowOffset = rowOffset + 1
      end if

      do idBand = 1, 9
      if (.not. instrBandPresent(idInst, idBand)) cycle
        write(oldColName, '(a3, i1, a4)') instrStr(idInst)//'_', idBand, '_EXP'
        oldExpMap => real32Data(column(srcmatchTab, oldColName))

        do srcNum = 1, numSources
          row = numInstsBandsPerSource * (srcNum-1) + rowOffset
          expMap(row) = oldExpMap(srcNum-1)
        end do
        rowOffset = rowOffset + 1
      end do ! idband
    end do ! idInst
    call release(expMapCol)

    ! FLUX
if (testFlag) print *, 'FLUX'

    ! First deal with rows for which ID_INST==0:
    rowOffset = 0
    if (instrBandPresent(0, 0)) then
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1) + rowOffset
        call setToNull(fluxCol, row)
      end do
      rowOffset = rowOffset + 1
    end if

!    flux => real32Data(fluxCol)
!    do idInst = 1, 3
!      if (instrBandPresent(idInst, 0)) then
!        oldColName = instrStr(idInst)//'_TOT_FLUX'
!        oldFlux => real32Data(column(srcmatchTab, oldColName))
!        do srcNum = 1, numSources
!          row = numInstsBandsPerSource * (srcNum-1) + rowOffset
!          flux(row) = oldFlux(srcNum-1)
!        end do
!        rowOffset = rowOffset + 1
!      end if
!
!      do idBand = 1, 9
!      if (.not. instrBandPresent(idInst, idBand)) cycle
!        write(oldColName, '(a3, i1, a5)') instrStr(idInst)//'_', idBand, '_FLUX'
!        oldFlux => real32Data(column(srcmatchTab, oldColName))
!
!        do srcNum = 1, numSources
!          row = numInstsBandsPerSource * (srcNum-1) + rowOffset
!          flux(row) = oldFlux(srcNum-1)
!        end do
!        rowOffset = rowOffset + 1
!      end do ! idband
!    end do ! idInst
    call release(fluxCol)

    ! FLUX_ERR
if (testFlag) print *, 'FLUX_ERR'

    ! First deal with rows for which ID_INST==0:
    rowOffset = 0
    if (instrBandPresent(0, 0)) then
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1) + rowOffset
        call setToNull(fluxErrCol, row)
      end do
      rowOffset = rowOffset + 1
    end if

!    fluxErr => real32Data(fluxErrCol)
!    do idInst = 1, 3
!      if (instrBandPresent(idInst, 0)) then
!        oldColName = 'ERR_'//instrStr(idInst)//'_TOT_FLUX'
!        oldFluxErr => real32Data(column(srcmatchTab, oldColName))
!        do srcNum = 1, numSources
!          row = numInstsBandsPerSource * (srcNum-1) + rowOffset
!          fluxErr(row) = oldFluxErr(srcNum-1)
!        end do
!        rowOffset = rowOffset + 1
!      end if
!
!      do idBand = 1, 9
!      if (.not. instrBandPresent(idInst, idBand)) cycle
!        write(oldColName, '(a7, i1, a5)')&
!          'ERR_'//instrStr(idInst)//'_', idBand, '_FLUX'
!        oldFluxErr => real32Data(column(srcmatchTab, oldColName))
!
!        do srcNum = 1, numSources
!          row = numInstsBandsPerSource * (srcNum-1) + rowOffset
!          fluxErr(row) = oldFluxErr(srcNum-1)
!        end do
!        rowOffset = rowOffset + 1
!      end do ! idband
!    end do ! idInst
    call release(fluxErrCol)

    ! RATE
if (testFlag) print *, 'RATE'

    rate => real32Data(rateCol)
    ! First deal with rows for which ID_INST==0:
    rowOffset = 0
    if (instrBandPresent(0, 0)) then
      oldColName = 'EP_TOT'
      oldRate => real32Data(column(srcmatchTab, oldColName))
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1) + rowOffset
        rate(row) = oldRate(srcNum-1)
      end do
      rowOffset = rowOffset + 1
    end if

    do idInst = 1, 3
      if (instrBandPresent(idInst, 0)) then
        oldColName = instrStr(idInst)//'_TOT'
        oldRate => real32Data(column(srcmatchTab, oldColName))
        do srcNum = 1, numSources
          row = numInstsBandsPerSource * (srcNum-1) + rowOffset
          rate(row) = oldRate(srcNum-1)
        end do
        rowOffset = rowOffset + 1
      end if

      do idBand = 1, 9
      if (.not. instrBandPresent(idInst, idBand)) cycle
        write(oldColName, '(a3, i1)') instrStr(idInst)//'_', idBand
        oldRate => real32Data(column(srcmatchTab, oldColName))

        do srcNum = 1, numSources
          row = numInstsBandsPerSource * (srcNum-1) + rowOffset
          rate(row) = oldRate(srcNum-1)
        end do
        rowOffset = rowOffset + 1
      end do ! idband
    end do ! idInst
    call release(rateCol)

    ! RATE_ERR
if (testFlag) print *, 'RATE_ERR'

    rateErr => real32Data(rateErrCol)
    ! First deal with rows for which ID_INST==0:
    rowOffset = 0
    if (instrBandPresent(0, 0)) then
      oldColName = 'EP_TOT_ERR'
      oldRateErr => real32Data(column(srcmatchTab, oldColName))
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1) + rowOffset
        rateErr(row) = oldRateErr(srcNum-1)
      end do
      rowOffset = rowOffset + 1
    end if

    do idInst = 1, 3
      if (instrBandPresent(idInst, 0)) then
        oldColName = instrStr(idInst)//'_TOT_ERR'
        oldRateErr => real32Data(column(srcmatchTab, oldColName))
        do srcNum = 1, numSources
          row = numInstsBandsPerSource * (srcNum-1) + rowOffset
          rateErr(row) = oldRateErr(srcNum-1)
        end do
        rowOffset = rowOffset + 1
      end if

      do idBand = 1, 9
      if (.not. instrBandPresent(idInst, idBand)) cycle
        write(oldColName, '(a3, i1, a4)') instrStr(idInst)//'_', idBand, '_ERR'
        oldRateErr => real32Data(column(srcmatchTab, oldColName))

        do srcNum = 1, numSources
          row = numInstsBandsPerSource * (srcNum-1) + rowOffset
          rateErr(row) = oldRateErr(srcNum-1)
        end do
        rowOffset = rowOffset + 1
      end do ! idband
    end do ! idInst
    call release(rateErrCol)

    ! RA
if (testFlag) print *, 'RA'

    ra => real64Data(raCol)
    oldRa => real64Data(column(srcmatchTab, 'RA'))
    do srcNum = 1, numSources
      startRow = numInstsBandsPerSource * (srcNum-1)
      finisRow = numInstsBandsPerSource * srcNum - 1
      ra(startRow:finisRow) = oldRa(srcNum-1)
    end do
    call release(raCol)

    ! DEC
if (testFlag) print *, 'DEC'

    dec => real64Data(decCol)
    oldDec => real64Data(column(srcmatchTab, 'DEC'))
    do srcNum = 1, numSources
      startRow = numInstsBandsPerSource * (srcNum-1)
      finisRow = numInstsBandsPerSource * srcNum - 1
      dec(startRow:finisRow) = oldDec(srcNum-1)
    end do
    call release(decCol)

    ! RADEC_ERR
if (testFlag) print *, 'RADEC_ERR'

    raDecErr => real32Data(raDecErrCol)
    oldRaDecErr => real32Data(column(srcmatchTab, 'RADEC_ERR'))
    do srcNum = 1, numSources
      startRow = numInstsBandsPerSource * (srcNum-1)
      finisRow = numInstsBandsPerSource * srcNum - 1
      raDecErr(startRow:finisRow) = oldRaDecErr(srcNum-1)
    end do
    call release(raDecErrCol)

    ! LII
if (testFlag) print *, 'LII'

    lii => real64Data(liiCol)
    oldLii => real64Data(column(srcmatchTab, 'LII'))
    do srcNum = 1, numSources
      startRow = numInstsBandsPerSource * (srcNum-1)
      finisRow = numInstsBandsPerSource * srcNum - 1
      lii(startRow:finisRow) = oldLii(srcNum-1)
    end do
    call release(liiCol)

    ! BII
if (testFlag) print *, 'BII'

    bii => real64Data(biiCol)
    oldBii => real64Data(column(srcmatchTab, 'BII'))
    do srcNum = 1, numSources
      startRow = numInstsBandsPerSource * (srcNum-1)
      finisRow = numInstsBandsPerSource * srcNum - 1
      bii(startRow:finisRow) = oldBii(srcNum-1)
    end do
    call release(biiCol)

    ! RAWX
if (testFlag) print *, 'RAWX'

    call setNullValue(rawXCol, -1)
    do row = 0, numRowsEmldetectList-1
      call setToNull(rawXCol, row)
    end do
    call release(rawXCol)

    ! RAWY
if (testFlag) print *, 'RAWY'

    call setNullValue(rawYCol, -1)
    do row = 0, numRowsEmldetectList-1
      call setToNull(rawYCol, row)
    end do
    call release(rawYCol)

    ! CCDNR
if (testFlag) print *, 'CCDNR'

    call setNullValue(ccdnrCol, -1)
    do row = 0, numRowsEmldetectList-1
      call setToNull(ccdnrCol, row)
    end do
    call release(ccdnrCol)

    ! HR1
if (testFlag) print *, 'HR1'

    hr1 => real32Data(hr1Col)
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      firstNonInst0Row = 1
      oldColName = 'EP_HR1'
      oldHr1 => real32Data(column(srcmatchTab, oldColName))
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        hr1(row) = oldHr1(srcNum-1)
      end do
    end if

    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      oldColName = instrStr(idInst)//'_HR1'
      oldHr1 => real32Data(column(srcmatchTab, oldColName))
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        hr1(startRow:finisRow) = oldHr1(srcNum-1)
      end do
    end do
    call release(hr1Col)

    ! HR1_ERR
if (testFlag) print *, 'HR1_ERR'

    hr1Err => real32Data(hr1ErrCol)
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      firstNonInst0Row = 1
      oldColName = 'EP_HR1_ERR'
      oldHr1Err => real32Data(column(srcmatchTab, oldColName))
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        hr1Err(row) = oldHr1Err(srcNum-1)
      end do
    end if

    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      oldColName = instrStr(idInst)//'_HR1_ERR'
      oldHr1Err => real32Data(column(srcmatchTab, oldColName))
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        hr1Err(startRow:finisRow) = oldHr1Err(srcNum-1)
      end do
    end do
    call release(hr1ErrCol)

    ! HR2
if (testFlag) print *, 'HR2'

    hr2 => real32Data(hr2Col)
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      firstNonInst0Row = 1
      oldColName = 'EP_HR2'
      oldHr2 => real32Data(column(srcmatchTab, oldColName))
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        hr2(row) = oldHr2(srcNum-1)
      end do
    end if

    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      oldColName = instrStr(idInst)//'_HR2'
      oldHr2 => real32Data(column(srcmatchTab, oldColName))
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        hr2(startRow:finisRow) = oldHr2(srcNum-1)
      end do
    end do
    call release(hr2Col)

    ! HR2_ERR
if (testFlag) print *, 'HR2_ERR'

    hr2Err => real32Data(hr2ErrCol)
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      firstNonInst0Row = 1
      oldColName = 'EP_HR2_ERR'
      oldHr2Err => real32Data(column(srcmatchTab, oldColName))
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        hr2Err(row) = oldHr2Err(srcNum-1)
      end do
    end if

    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      oldColName = instrStr(idInst)//'_HR2_ERR'
      oldHr2Err => real32Data(column(srcmatchTab, oldColName))
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        hr2Err(startRow:finisRow) = oldHr2Err(srcNum-1)
      end do
    end do
    call release(hr2ErrCol)

    ! HR3
if (testFlag) print *, 'HR3'

    hr3 => real32Data(hr3Col)
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      firstNonInst0Row = 1
      oldColName = 'EP_HR3'
      oldHr3 => real32Data(column(srcmatchTab, oldColName))
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        hr3(row) = oldHr3(srcNum-1)
      end do
    end if

    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      oldColName = instrStr(idInst)//'_HR3'
      oldHr3 => real32Data(column(srcmatchTab, oldColName))
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        hr3(startRow:finisRow) = oldHr3(srcNum-1)
      end do
    end do
    call release(hr3Col)

    ! HR3_ERR
if (testFlag) print *, 'HR3_ERR'

    hr3Err => real32Data(hr3ErrCol)
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      firstNonInst0Row = 1
      oldColName = 'EP_HR3_ERR'
      oldHr3Err => real32Data(column(srcmatchTab, oldColName))
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        hr3Err(row) = oldHr3Err(srcNum-1)
      end do
    end if

    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      oldColName = instrStr(idInst)//'_HR3_ERR'
      oldHr3Err => real32Data(column(srcmatchTab, oldColName))
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        hr3Err(startRow:finisRow) = oldHr3Err(srcNum-1)
      end do
    end do
    call release(hr3ErrCol)

    ! CUTRAD
if (testFlag) print *, 'CUTRAD'

    do row = 0, numRowsEmldetectList-1
      call setToNull(cutRadCol, row)
    end do
    call release(cutRadCol)

    ! MASKFRAC
if (testFlag) print *, 'MASKFRAC'

    do row = 0, numRowsEmldetectList-1
      call setToNull(maskFracCol, row)
    end do
    call release(maskFracCol)

    ! EEF
if (testFlag) print *, 'EEF'

    do row = 0, numRowsEmldetectList-1
      call setToNull(eefCol, row)
    end do
    call release(eefCol)

    ! VIGNETTING
if (testFlag) print *, 'VIGNETTING'

    do row = 0, numRowsEmldetectList-1
      call setToNull(vignetCol, row)
    end do
    call release(vignetCol)

    ! DIST_NN
if (testFlag) print *, 'DIST_NN'

    do row = 0, numRowsEmldetectList-1
      call setToNull(distNNCol, row)
    end do
    call release(distNNCol)

    ! VAL_FLAG
if (testFlag) print *, 'VAL_FLAG'

    emptyStr = ''
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      firstNonInst0Row = 1
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        call setStringCell(valFlagCol, row, emptyStr)
      end do
    end if

    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      oldColName = instrStr(idInst)//'_VAL_FLAG'
      oldValFlagCol = column(srcmatchTab, oldColName)
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        do row = startRow, finisRow
          call setStringCell(valFlagCol, row&
            , stringCell(oldValFlagCol, srcNum-1))
        end do
      end do
      call release(oldValFlagCol)
    end do
    call release(valFlagCol)

    ! VER_FLAG
if (testFlag) print *, 'VER_FLAG'

    emptyStr = ''
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      firstNonInst0Row = 1
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        call setStringCell(verFlagCol, row, emptyStr)
      end do
    end if

    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      oldColName = instrStr(idInst)//'_VER_FLAG'
      oldVerFlagCol = column(srcmatchTab, oldColName)
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        do row = startRow, finisRow
          call setStringCell(verFlagCol, row&
            , stringCell(oldVerFlagCol, srcNum-1))
        end do
      end do
      call release(oldVerFlagCol)
    end do
    call release(verFlagCol)

    ! VER_COMM
if (testFlag) print *, 'VER_COMM'

    emptyStr = ''
    firstNonInst0Row = 0
    if (instrBandPresent(0, 0)) then
      firstNonInst0Row = 1
      do srcNum = 1, numSources
        row = numInstsBandsPerSource * (srcNum-1)
        call setStringCell(verCommCol, row, emptyStr)
      end do
    end if

    instFinisOffset = firstNonInst0Row - 1
    do idInst = 1, 3
      instStartOffset = instFinisOffset + 1
      instFinisOffset = instStartOffset + count(instrBandPresent(idInst, :)) - 1
      oldColName = instrStr(idInst)//'_VER_COMM'
      oldVerCommCol = column(srcmatchTab, oldColName)
      do srcNum = 1, numSources
        startRow = numInstsBandsPerSource * (srcNum-1) + instStartOffset
        finisRow = numInstsBandsPerSource * (srcNum-1) + instFinisOffset
        do row = startRow, finisRow
          call setStringCell(verCommCol, row&
            , stringCell(oldVerCommCol, srcNum-1))
        end do
      end do
      call release(oldVerCommCol)
    end do
    call release(verCommCol)

    !...................................................................

    call release(emldetectSet)
    call release(srcmatchSet)
if (testFlag) print *, 'zzz'

  end subroutine convertFromSrcmatchToEmldetect

  !---------------------------------------------------------------------
  subroutine addBoxId
  !---------------------------------------------------------------------
  ! Adds column BOX_ID_SRC to the output of convertFromSrcmatchToEmldetect.

    implicit none

    type(DataSetT) :: convSet, mlSet
    type(TableT) :: convTab, mlTab
    integer :: numRowsMl, numRowsConv, mlRow
    integer(int32), pointer ::&
      mlSrcId(:), mlBoxId(:), mlIdInst(:), mlIdBand(:),&
      convSrcId(:), convIdInst(:), convIdBand(:), convBoxId(:)
    logical, allocatable :: mask(:)
    type(ColumnT) :: convBoxIdCol, ximaCol, yimaCol
    type(WcsT) :: wcs

    !-------------------------------------------------------------------
    ! Start

    mlSet = dataSet(inSetName, READ)
    mlTab = table(mlSet, 'SRCLIST')
    numRowsMl = numberOfRows(mlTab)

    mlSrcId  => int32Data(column(mlTab, 'ML_ID_SRC'))
    mlBoxId  => int32Data(column(mlTab, 'BOX_ID_SRC'))
    mlIdInst => int32Data(column(mlTab, 'ID_INST'))
    mlIdBand => int32Data(column(mlTab, 'ID_BAND'))

    convSet = dataSet(outSetName, MODIFY)
    convTab = table(convSet, 'SRCLIST')
    numRowsConv = numberOfRows(convTab)
    allocate(mask(numRowsConv))
    convSrcId  => int32Data(column(convTab, 'ML_ID_SRC'))
    convIdInst => int32Data(column(convTab, 'ID_INST'))
    convIdBand => int32Data(column(convTab, 'ID_BAND'))

    convBoxIdCol = addOrOpenColumn(convTab, 'BOX_ID_SRC', INTEGER32)
    ! addOrOpenColumn is in ssclib/dal_aux.

    convBoxId => int32Data(convBoxIdCol)

    do mlRow = 0, numRowsMl - 1
      mask = (convSrcId  == mlSrcId( mlRow)&
        .and. convIdInst == mlIdInst(mlRow)&
        .and. convIdBand == mlIdBand(mlRow))

      where(mask) convBoxId = mlBoxId(mlRow)
    end do

    call readWcsSourceList(mlSet, wcs) ! in ssclib/wcs_aux.
    ximaCol = column(convTab, 'X_IMA')
    yimaCol = column(convTab, 'Y_IMA')
    if (.not. (hasAttribute(ximaCol, 'TCTYP')&
    .and.      hasAttribute(yimaCol, 'TCTYP')))&
      call writeWcs(ximaCol, yimaCol, wcs) ! in ssclib/wcs_aux.

    call release(convSet)
    call release(mlSet)

  end subroutine addBoxId

  !---------------------------------------------------------------------
  subroutine sky2Pix
  !---------------------------------------------------------------------
  ! Adds column BOX_ID_SRC to the output of convertFromSrcmatchToEmldetect.

    implicit none

    type(DataSetT) :: convSet
    type(TableT) :: convTab
    integer :: numRowsConv, srcNum, row
    logical, allocatable :: mask(:)
    integer(int32), pointer :: convSrcId(:)
    real(single), pointer, dimension(:) ::&
      convRaDecErrArcsec, convXIma, convXImaErr, convYIma, convYImaErr
    real(double), pointer :: convRaDeg(:), convDecDeg(:)
    type(ColumnT) :: ximaCol, yimaCol
    type(WcsT) :: wcs
    real(double) ::&
      xPixelSizeWorld, yPixelSizeWorld, rotAngle, ra, dec, refRa, refDec,&
      xTan, yTan
    real(single) :: xIma, yIma, xImaErr, yImaErr

    !-------------------------------------------------------------------
    ! Start

    convSet = dataSet(outSetName, MODIFY)
    convTab = table(convSet, 'SRCLIST')
    numRowsConv = numberOfRows(convTab)
    allocate(mask(numRowsConv))
    convSrcId          => int32Data(column(convTab, 'SRC_NUM'))
    convRaDeg          => real64Data(column(convTab, 'RA'))
    convDecDeg         => real64Data(column(convTab, 'DEC'))
    convRaDecErrArcsec => real32Data(column(convTab, 'RADEC_ERR'))
    ximaCol = column(convTab, 'X_IMA')
    convXIma           => real32Data(ximaCol)
    convXImaErr        => real32Data(column(convTab, 'X_IMA_ERR'))
    yimaCol = column(convTab, 'Y_IMA')
    convYIma           => real32Data(yimaCol)
    convYImaErr        => real32Data(column(convTab, 'Y_IMA_ERR'))

    call readWcs(ximaCol, yimaCol, wcs) ! in ssclib/wcs_aux.
    call convertNewToOld(wcs, xPixelSizeWorld, yPixelSizeWorld, rotAngle)

    do srcNum = minval(convSrcId), maxval(convSrcId)
      mask = (convSrcId == srcNum)
    if (.not. any(mask)) cycle
      row = smallestTrueIndex(mask) ! in ssclib/array_utils.

      ra  = convRaDeg( row-1) * PiOn180
      dec = convDecDeg(row-1) * PiOn180
      refRa  = wcs%x%refWorld * PiOn180
      refDec = wcs%y%refWorld * PiOn180
      call raDecToTan(ra, dec, refRa, refDec, xTan, yTan) ! in ssclib/coordinate
      call tanToPixels(wcs, xTan, yTan, xIma, yIma) ! in ssclib/wcs_aux.
      xImaErr = convRaDecErrArcsec(row-1) / 3600.0 / abs(xPixelSizeWorld)
      yImaErr = convRaDecErrArcsec(row-1) / 3600.0 / abs(yPixelSizeWorld)

      where(mask)
        convXIma    = xIma
        convXImaErr = xImaErr
        convYIma    = yIma
        convYImaErr = yImaErr
      end where
    end do

    call release(convSet)

  end subroutine sky2Pix

  !---------------------------------------------------------------------
  subroutine srcListPatch
  !---------------------------------------------------------------------
  ! Adds columns SRC_BOX_CTS, SRC_BOX_CTS_ERR and BKG_BOX_CTS to the emldetect (or srcmatch) output (taking the eboxdetect output as input) and also copies the LIKE column from the eboxdetect file.

    implicit none

    type(DataSetT) :: boxSet, mlSet
    type(TableT) :: boxTab, mlTab
    type(ColumnT) ::&
      boxIndexCol, boxInstrumCol, boxIdBandCol, boxBoxCtsCol,&
      boxBgMapCol, boxBoxSizeCol, boxLikeCol, mlLikeCol,&
      mlBoxSizeCol, mlBkgBoxCtsCol, mlSrcBoxCtsCol,&
      mlSrcBoxCtsErrCol, mlBoxIndexCol, mlInstrumCol, mlIdbandCol
    integer(int32), pointer ::&
      boxIndex(:), boxInstrum(:), boxIdband(:), mlBoxIndex(:),&
      mlInstrum(:), mlIdband(:)
    real(single), pointer, dimension(:) ::&
      boxBoxCts, boxBgMap, boxBoxSize, boxLike, mlLike, mlBkgBoxCts,&
      mlSrcBoxCts, mlSrcBoxCtsErr, mlBoxSize
    integer :: numBoxSrcs, numMlRows, mlRow, boxRow
    logical(bool), allocatable :: mask(:)

    !-------------------------------------------------------------------
    ! Start

    boxSet = dataSet(inSetName, READ)
    boxTab = table(boxSet, 'SRCLIST')
    numBoxSrcs = numberOfRows(boxTab)
    allocate(mask(numBoxSrcs))

    boxIndexCol   = column(boxTab, 'BOX_ID_SRC')
    boxInstrumCol = column(boxTab, 'ID_INST')
    boxIdBandCol  = column(boxTab, 'ID_BAND')
    boxBoxCtsCol  = column(boxTab, 'BOX_CTS')
    boxBgMapCol   = column(boxTab, 'BG_MAP')
    boxBoxSizeCol = column(boxTab, 'BOX_SIZE')
    boxLikeCol    = column(boxTab, 'LIKE')
    boxIndex   => int32Data( boxIndexCol)
    boxInstrum => int32Data( boxInstrumCol)
    boxIdband  => int32Data( boxIdBandCol)
    boxBoxCts  => real32Data(boxBoxCtsCol)
    boxBgMap   => real32Data(boxBgMapCol)
    boxBoxSize => real32Data(boxBoxSizeCol)
    boxLike    => real32Data(boxLikeCol)

    mlSet = dataSet(outSetName, MODIFY)
    mlTab = table(mlSet, 'SRCLIST')
    numMlRows = numberOfRows(mlTab)
!    allocate(mask(numMlRows))
    mlBoxIndexCol = column(mlTab, 'BOX_ID_SRC')
    mlInstrumCol  = column(mlTab, 'ID_INST')
    mlIdbandCol   = column(mlTab, 'ID_BAND')
!    if (.not. nullDefined(mlBoxIndexCol)) then
!      call warning('noNullDefinedBoxIdCol',"")
!      call setNullValue(mlBoxIndexCol, -1)
!    end if
    mlBoxIndex => int32Data(mlBoxIndexCol)
    mlInstrum  => int32Data(mlInstrumCol)
    mlIdband   => int32Data(mlIdbandCol)

    mlLikeCol         = addOrOpenColumn(mlTab, 'LIKE'&
      , REAL32, '',       'eboxdetect detection likelihood.')
    mlBoxSizeCol      = addOrOpenColumn(mlTab, 'BOX_SIZE'&
      , REAL32, 'pixels', 'eboxdetect square box dimension.')
    mlBkgBoxCtsCol    = addOrOpenColumn(mlTab, 'BKG_BOX_CTS'&
      , REAL32, 'counts', 'BG_MAP * BOX_SIZE * BOX_SIZE')
    mlSrcBoxCtsCol    = addOrOpenColumn(mlTab, 'SRC_BOX_CTS'&
      , REAL32, 'counts', 'BOX_CTS - BKG_BOX_CTS')
    mlSrcBoxCtsErrCol = addOrOpenColumn(mlTab, 'SRC_BOX_CTS_ERR'&
      , REAL32, 'counts', 'sqrt(BOX_CTS)')
! - BKG_BOX_CTS')
    ! addOrOpenColumn is in ssclib/dal_aux.

!    do mlRow = 0, numMlRows - 1
!      call setToNull(mlLikeCol,         mlRow)
!      call setToNull(mlBkgBoxCtsCol,    mlRow)
!      call setToNull(mlSrcBoxCtsCol,    mlRow)
!      call setToNull(mlSrcBoxCtsErrCol, mlRow)
!    end do
!*** no need - its already been done when the ml list was constructed, and if you do it again here you can overwrite values for previous instruments.

    mlLike         => real32Data(mlLikeCol)
    mlBoxSize      => real32Data(mlBoxSizeCol)
    mlBkgBoxCts    => real32Data(mlBkgBoxCtsCol)
    mlSrcBoxCts    => real32Data(mlSrcBoxCtsCol)
    mlSrcBoxCtsErr => real32Data(mlSrcBoxCtsErrCol)

    do mlRow = 0, numMlRows - 1
!print *, 'mlRow=', mlRow
    if (mlIdband(mlRow) == 9) cycle ! don't want the xid band.
    if (isNull(mlBoxIndexCol, mlRow)) cycle

      mask = (boxIndex   == mlBoxIndex(mlRow)&
        .and. boxInstrum == mlInstrum(mlRow)&
        .and. boxIdband  == mlIdband(mlRow))
if (count(mask) .ne. 1) then
!print *, 'BOX_ID_SRC=', mlBoxIndex(mlRow)
!print *, '   ID_INST=', mlInstrum(mlRow)
!print *, '   ID_BAND=', mlIdband(mlRow)
!print *, 'count(mask)=', count(mask)
!call error('boxIndexNotFound',"")
cycle
end if

      boxRow = smallestTrueIndex(mask) - 1

      mlLike(mlRow) = boxLike(boxRow)
      mlBoxSize(mlRow) = boxBoxSize(boxRow)
      mlBkgBoxCts(mlRow) = boxBgMap(boxRow) * boxBoxSize(boxRow)&
        * boxBoxSize(boxRow)
      mlSrcBoxCts(mlRow) = boxBoxCts(boxRow) - mlBkgBoxCts(mlRow)
    if (mlSrcBoxCts(mlRow) < 0.0) cycle ! leave the err at null.
      mlSrcBoxCtsErr(mlRow) = sqrt(boxBoxCts(boxRow))
! - mlBkgBoxCts(mlRow))
    end do

    call release(mlSrcBoxCtsErrCol)
    call release(mlSrcBoxCtsCol)
    call release(mlBkgBoxCtsCol)
    call release(mlBoxSizeCol)
    call release(mlLikeCol)
    call release(mlIdbandCol)
    call release(mlInstrumCol)
    call release(mlBoxIndexCol)
    call release(mlTab)
    call release(mlSet)

    call release(boxLikeCol)
    call release(boxIndexCol)
    call release(boxInstrumCol)
    call release(boxIdBandCol)
    call release(boxBoxCtsCol)
    call release(boxBgMapCol)
    call release(boxBoxSizeCol)
    call release(boxTab)
    call release(boxSet)

  end subroutine srcListPatch

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module edetaux_mod
