! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: l_dal_aux_2
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some routines for dealing with keywords and columns of FITS files.
! 
! Comments: TODO:
!*********** Look through ALL software for unnecessary pointer allocations!
!*********** Look through ALL software and declare all ptrs => null()!
! 
! $Id: l_dal_aux_2.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- fred
module l_dal_aux_2

  use dal
  use errorHandling

  use constants	! PiOn2
  use array_utils
  use sort_mod
!use coordinate
use dump

  implicit none 

  private 

! dal_aux:
  public l_addOrOpenColumn
  public myAddColumn
  public readArrayDataName2dLogical
  public readArrayDataName3dLogical

  interface myAddColumn
    module procedure myAddColumnInt32
    module procedure myAddColumnSingle
    module procedure myAddColumnDouble
  end interface

  contains

  !---------------------------------------------------------------------
  subroutine readArrayDataName2dLogical(imageSetName, image)
  !---------------------------------------------------------------------
!*** put it in dal_aux to replace the bool one!

  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    character(*), intent(in) :: imageSetName
    logical,      pointer    :: image(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call readArrayDataArray2dLogical(inArray, image)
    call release(set)

  end subroutine readArrayDataName2dLogical

  !---------------------------------------------------------------------
  subroutine readArrayDataArray2dLogical(inArray, image)
  !---------------------------------------------------------------------
!*** put it in dal_aux to replace the bool one!

  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    type(ArrayT), intent(in) :: inArray
    logical, dimension(:,:), pointer :: image

    integer(int32), pointer :: dimVector(:)
    logical, pointer :: cube(:,:,:)
    integer ::&
      numDimensions, arrayShape(3), biggestDimensions(2, 1), imageShape(2)
    logical :: mask(3)

    integer(int8),  pointer :: arrayDataInt8(  :,:)
    integer(int16), pointer :: arrayDataInt16( :,:)
    integer(int32), pointer :: arrayDataInt32( :,:)
    real(single),   pointer :: arrayDataReal32(:,:)
    real(double),   pointer :: arrayDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)
    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(image(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        image = (arrayDataInt8 > 0)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        image = (arrayDataInt16 > 0)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        image = (arrayDataInt32 > 0)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        image = (arrayDataReal32 > 0)
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        image = (arrayDataReal64 > 0)
      end if

    case(3)
      call readArrayDataArray3dLogical(inArray, cube)
      arrayShape(:3) = shape(cube)
if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      imageShape(1) = size(cube, biggestDimensions(1,1))
      imageShape(2) = size(cube, biggestDimensions(2,1))
      allocate(image(imageShape(1), imageShape(2)))
      image = reshape(cube, imageShape)
      deallocate(cube)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine readArrayDataArray2dLogical

  !---------------------------------------------------------------------
  subroutine readArrayDataName3dLogical(imageSetName, cube)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate cube() after use.

    implicit none

    character(*), intent(in) :: imageSetName
    logical,      pointer    :: cube(:,:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call readArrayDataArray3dLogical(inArray, cube)
    call release(set)

  end subroutine readArrayDataName3dLogical

  !---------------------------------------------------------------------
  subroutine readArrayDataArray3dLogical(inArray, cube)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate cube() after use.

    implicit none

    type(ArrayT), intent(in) :: inArray
    logical,      pointer    :: cube(:,:,:)

    integer(int32), pointer :: dimVector(:)
    integer :: numDimensions
    integer(int8),  pointer :: arrayDataInt8(  :,:,:)
    integer(int16), pointer :: arrayDataInt16( :,:,:)
    integer(int32), pointer :: arrayDataInt32( :,:,:)
    real(single),   pointer :: arrayDataReal32(:,:,:)
    real(double),   pointer :: arrayDataReal64(:,:,:)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)

    select case(numDimensions)
    case(1, 2)
call error('arrayRankTooSmall', '')
    case(3)
      allocate(cube(dimVector(0), dimVector(1), dimVector(2)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array3Data(inArray)
        cube = (arrayDataInt8 > 0)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array3Data(inArray)
        cube = (arrayDataInt16 > 0)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array3Data(inArray)
        cube = (arrayDataInt32 > 0)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array3Data(inArray)
        cube = (arrayDataReal32 > 0)
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array3Data(inArray)
        cube = (arrayDataReal64 > 0)
      end if

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine readArrayDataArray3dLogical

  !---------------------------------------------------------------------
  function l_addOrOpenColumn(tab, colName, inDataType, units, comment&
  , integerNullValue) result(col)
  !---------------------------------------------------------------------
!*** export modifications to dal_aux??

    implicit none

    type(TableT),   intent(in)           :: tab
    character(*),   intent(in)           :: colName
    integer(int32), intent(in), optional :: inDataType, integerNullValue
    character(*),   intent(in), optional :: units, comment
    type(ColumnT) :: col

    integer(int32) :: localDataType, localNullValue, row
    character(255) :: localUnits, localComment

    !-------------------------------------------------------------------
    ! Start

    if (present(inDataType)) then
      localDataType = inDataType
    else
      localDataType = REAL32
    end if

    if (present(units)) then
      localUnits = units
    else
      localUnits = ''
    end if

    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if

    if (present(integerNullValue)) then
      localNullValue = integerNullValue
    else
      localNullValue = -99
    end if

    if (hasColumn(tab, colName)) then
      col = column(tab, colName)

      if (present(inDataType) .and. dataType(col) .ne. localDataType) then
        call release(col)
        call deleteColumn(tab, colName)
        if (.not. present(inDataType)) call warning('noDataType',"The &
          &specified column doesn't exist in the file, so I am adding it. &
          &You didn't specify a data type so I'm making it REAL32. Hope &
          &that's ok!")

        col = addColumn(tab, colName, localDataType, localUnits&
          , comment=localComment)
      end if

    else
      if (.not. present(inDataType)) call warning('noDataType',"The specified &
        &column doesn't exist in the file, so I am adding it. You didn't &
        &specify a data type so I'm making it REAL32. Hope that's ok!")

      col = addColumn(tab, colName, localDataType, localUnits&
        , comment=localComment)

      select case(localDataType)
      case(INTEGER8)
        call setNullValue(col, int(int(localNullValue, int8), int32))
      case(INTEGER16)
        call setNullValue(col, int(int(localNullValue, int16), int32))
      case(INTEGER32)
        call setNullValue(col, localNullValue)
      end select

      do row = 0, numberOfRows(col)-1
        call setToNull(col, row)
      end do
    end if

  end function l_addOrOpenColumn

  !---------------------------------------------------------------------
  subroutine myAddColumnInt32(tab, name, inData, unit, comment, nulls&
  , nullValue)
  !---------------------------------------------------------------------

!*********** OBSOLETE - function taken over by writeColumn.

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: name
    character(*), intent(in), optional :: unit, comment
    logical, intent(in), optional :: nulls(:)
    integer(int32), intent(in), optional :: nullValue
    integer(int32), intent(in) :: inData(:)

    character(255) :: localUnit, localComment
    logical :: localNulls(size(inData))
    type(ColumnT) :: col
    integer(int32), pointer :: colData(:)
    integer(int32) :: i, localNullValue

    !-------------------------------------------------------------------
    ! Start

    if (present(unit)) then
      localUnit = unit
    else
      localUnit = ''
    end if
    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if
    if (present(nullValue)) then
      localNullValue = nullValue
    else
      localNullValue = -999
    end if
    if (present(nulls)) then
      localNulls = .true.
      if (size(nulls) .eq. size(inData)) then
        localNulls = nulls
      else if (size(nulls) .gt. size(inData)) then
        localNulls = nulls(1:size(inData))
      else
        localNulls(1:size(nulls)) = nulls
      end if
    else
      localNulls = .false.
    end if

    col = addColumn(tab, name, INTEGER32, trim(localUnit)&
    , comment=trim(localComment))

    colData => int32Data(col)
    colData = inData(1:size(inData))

    if (any(localNulls)) then
      call setNullValue(col, localNullValue)
      do i = 1, size(inData)
        if (localNulls(i)) call setToNull(col, i - 1)
      end do
    end if

    call release(col)

  end subroutine myAddColumnInt32

  !---------------------------------------------------------------------
  subroutine myAddColumnSingle(tab, name, inData, unit, comment, nulls)
  !---------------------------------------------------------------------

!*********** OBSOLETE - function taken over by writeColumn.


    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: name
    character(*), intent(in), optional :: unit, comment
    logical, intent(in), optional :: nulls(:)
    real(single), intent(in) :: inData(:)

    character(255) :: localUnit, localComment
    logical :: localNulls(size(inData))
    type(ColumnT) :: col
    real(single), pointer :: colData(:)
    integer(int32) :: i

    !-------------------------------------------------------------------
    ! Start

    if (present(unit)) then
      localUnit = unit
    else
      localUnit = ''
    end if
    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if
    if (present(nulls)) then
      localNulls = .true.
      if (size(nulls) .eq. size(inData)) then
        localNulls = nulls
      else if (size(nulls) .gt. size(inData)) then
        localNulls = nulls(1:size(inData))
      else
        localNulls(1:size(nulls)) = nulls
      end if
    else
      localNulls = .false.
    end if

    col = addColumn(tab, name, REAL32, trim(localUnit)&
    , comment=trim(localComment))

    colData => real32Data(col)
    colData = inData(1:size(inData))

    if (any(localNulls)) then
      do i = 1, size(inData)
        if (localNulls(i)) call setToNull(col, i - 1)
      end do
    end if

    call release(col)

  end subroutine myAddColumnSingle

  !---------------------------------------------------------------------
  subroutine myAddColumnDouble(tab, name, inData, unit, comment, nulls)
  !---------------------------------------------------------------------

!*********** OBSOLETE - function taken over by writeColumn.


    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: name
    character(*), intent(in), optional :: unit, comment
    logical, intent(in), optional :: nulls(:)
    real(double), intent(in) :: inData(:)

    character(255) :: localUnit, localComment
    logical :: localNulls(size(inData))
    type(ColumnT) :: col
    real(double), pointer :: colData(:)
    integer(int32) :: i

    !-------------------------------------------------------------------
    ! Start

    if (present(unit)) then
      localUnit = unit
    else
      localUnit = ''
    end if
    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if
    if (present(nulls)) then
      localNulls = .true.
      if (size(nulls) .eq. size(inData)) then
        localNulls = nulls
      else if (size(nulls) .gt. size(inData)) then
        localNulls = nulls(1:size(inData))
      else
        localNulls(1:size(nulls)) = nulls
      end if
    else
      localNulls = .false.
    end if

    col = addColumn(tab, name, REAL64, trim(localUnit)&
    , comment=trim(localComment))

    colData => real64Data(col)
    colData = inData(1:size(inData))

    if (any(localNulls)) then
      do i = 1, size(inData)
        if (localNulls(i)) call setToNull(col, i - 1)
      end do
    end if

    call release(col)

  end subroutine myAddColumnDouble

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module l_dal_aux_2
